
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#if defined(_OPENMP)
#include <omp.h>
#endif

/* compute high precision walltime and walltime difference */
static double wallclock(const double * ref)
{
    struct timespec t;
    double ret;

    clock_gettime(CLOCK_REALTIME, &t);
    ret = ((double) t.tv_sec) + 1.0e-9*((double) t.tv_nsec);

    return ref ? (ret - *ref) : ret;
}


/* extremely naiive matrix multiply C(n,o) = A(n,m) * B(m,o) */

void matmul_cpu(float *a, float *b, float *c, int n, int m, int o)
{
    int i,j,k;
    float sum;

    for (i = 0; i < n; ++i) {
        for (j = 0; j < o; ++j) {
            sum = 0.0f;
            for (k = 0; k < m; ++k) {
                sum += a[m*i+k] * b[o*k+j];
            }
            c[o*i+j] = sum;
        }
    }
}


/* improve performance by copying column data into a holding space 
 * and swapping loop order to have always dense data access */

void matmul_opt(float *x, float *y, float *z, int n, int m, int o)
    
{
}

int main(int argc, char **argv)
{
    double secs;
    float *a, *b, *c1, *c2;
    int n,m,o,i,flag;

    if (argc < 3) {
        printf("usage: %s <dim1> <dim2> <dim3>\n", argv[0]);
        return 1;
    }

    n = atoi(argv[1]);
    m = atoi(argv[2]);
    o = atoi(argv[3]);

    if (n < 1) {
        printf("dim1 must be > 1\n");
        return 2;
    }
    if (m < 1) {
        printf("dim2 must be > 1\n");
        return 3;
    }
    if (o < 1) {
        printf("dim3 must be > 1\n");
        return 4;
    }
    posix_memalign((void **)&a,64,sizeof(float)*n*m);
    posix_memalign((void **)&b,64,sizeof(float)*m*o);
    posix_memalign((void **)&c1,64,sizeof(float)*n*o);
    posix_memalign((void **)&c2,64,sizeof(float)*n*o);

    /* fill matrix a */
    for (i=0; i < n*m; ++i) {
        a[i] = rand()/(float)RAND_MAX;
    }
    /* fill matrix b */
    for (i=0; i < m*o; ++i) {
        b[i] = rand()/(float)RAND_MAX;
    }
    /* warm up cache */
    matmul_opt(a,b,c2,n,m,o);

    /* run simple CPU version and measure time */
    secs = wallclock(NULL);
    matmul_cpu(a,b,c1,n,m,o);
    printf("time for CPU code: %g s\n",wallclock(&secs));

    /* run optimized CPU version and measure time */
    secs = wallclock(NULL);
    matmul_opt(a,b,c2,n,m,o);
    printf("time for OPT code: %g s\n",wallclock(&secs));

    /* check result */
    flag=0;
    for (i=0; i < n*o; ++i) {
        if (fabsf((c1[i] - c2[i])/c1[i]) > 0.000002) {
            flag=1;
        }
    }
    if (flag == 0) {
        printf("OPT check OK\n");
    } else {
        printf("OPT check FAIL\n");
    }

    return 0;
}
